/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.namespace.QName;

import org.ow2.orchestra.cluster.ClusterDescription;
import org.ow2.orchestra.cluster.LocalServer;
import org.ow2.orchestra.cluster.Server;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.ManagementAPI;
import org.ow2.orchestra.facade.data.def.ProcessDefinitionData;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.OrchestraWrapperException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.jmx.commands.ActivateProcessCommand;
import org.ow2.orchestra.jmx.commands.DeployCommand;
import org.ow2.orchestra.jmx.commands.DeployWSCommand;
import org.ow2.orchestra.jmx.commands.GetActiveProcessDefinitionUUIDCommand;
import org.ow2.orchestra.jmx.commands.GetProcessCommand;
import org.ow2.orchestra.jmx.commands.GetProcessesCommand;
import org.ow2.orchestra.jmx.commands.InitializeCommand;
import org.ow2.orchestra.jmx.commands.RefreshProcessCommand;
import org.ow2.orchestra.jmx.commands.RemoveProcessCommand;
import org.ow2.orchestra.jmx.commands.RetireProcessCommand;
import org.ow2.orchestra.jmx.commands.UndeployCommand;
import org.ow2.orchestra.jmx.commands.UndeployWSCommand;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.cmd.Command;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.util.DeploymentUtil;
import org.ow2.orchestra.util.Misc;

public class ManagementAPIImpl implements ManagementAPI {

  private static final Logger LOG = Logger.getLogger(ManagementAPIImpl.class.getName());


  private final Server localServer;
  private final EnvironmentFactory environmentFactory;

  public ManagementAPIImpl(final EnvironmentFactory environmentFactory) {
    this.localServer = new LocalServer(ManagementAPIImpl.this);
    this.environmentFactory = environmentFactory;
  }

  public ManagementAPIImpl(final Server orchestraServer, final EnvironmentFactory environmentFactory) {
    this.localServer = orchestraServer;
    this.environmentFactory = environmentFactory;
  }

  protected CommandService getCommandService() {
    return this.environmentFactory.get(CommandService.class);
  }

  public <T> T execute(final Command<T> command) {
    return this.getCommandService().execute(command);
  }

  public ProcessDefinition deploy(final Deployment bpelDeployment) {
    final ProcessDefinition processDefinition = this.execute(new DeployCommand(bpelDeployment));
    if (processDefinition != null) {
      try {
        this.deployWebServicesOnCluster(processDefinition.getUUID(), false);
      } catch (final RuntimeException e) {
        this.undeployWebServicesOnCluster(processDefinition.getUUID(), true);
        this.execute(new UndeployCommand(processDefinition.getUUID()));
        throw e;
      }
      final Set<ProcessDefinitionUUID> otherVersions = this.getOtherVersionUUIDs(processDefinition.getUUID());
      for (final ProcessDefinitionUUID otherVersion : otherVersions) {
        this.refreshProcessOnCluster(otherVersion, true);
      }
      this.refreshProcessOnCluster(processDefinition.getUUID(), true);
    }
    return processDefinition;
  }

  private void deployWebServicesOnCluster(final ProcessDefinitionUUID processDefinitionUUID, final boolean ignoreExceptions) {
    // Get serverList
    final Collection<Server> servers = this.getOrchestraServersInCluster();
    // Deploy web services on each server
    for (final Server server : servers) {
      ManagementAPI managementAPI = null;
      if (server.equals(this.localServer)) {
        managementAPI = this;
      } else {
        // try to connect to server
        try {
          managementAPI = server.getManagementAPI();
        } catch (final Exception e) {
          // server is down, log a warning.
          ManagementAPIImpl.LOG.log(Level.WARNING, "Connection to server " + server + " failed.", e);
        }
      }
      // If server is up
      if (managementAPI != null) {
        try {
          // deploy web services on the server
          managementAPI.deployWebServices(processDefinitionUUID);
          ManagementAPIImpl.LOG.log(Level.INFO,
              "Web services for process " + processDefinitionUUID + " successfully deployed on server " + server + ".");
        } catch (final RuntimeException e) {
          // if exception
          if (ignoreExceptions) {
            ManagementAPIImpl.LOG.log(Level.WARNING,
                "Web services deployment for process " + processDefinitionUUID + " failed on server " + server + ".", e);
          } else {
            throw e;
          }
        }
      }
    }
  }

  private void undeployWebServicesOnCluster(final ProcessDefinitionUUID processDefinitionUUID, final boolean ignoreExceptions) {
    // Get serverList
    final Collection<Server> servers = this.getOrchestraServersInCluster();
    // Undeploy web services on each server
    for (final Server server : servers) {
      ManagementAPI managementAPI = null;
      // try to connect to server
      try {
        managementAPI = server.getManagementAPI();
      } catch (final Exception e) {
        // server is down, log a warning.
        ManagementAPIImpl.LOG.log(Level.WARNING, "Connection to server " + server + " failed.", e);
      }
      // If server is up
      if (managementAPI != null) {
        try {
          // undeploy web services on the server
          managementAPI.undeployWebServices(processDefinitionUUID);
          ManagementAPIImpl.LOG.log(Level.INFO,
              "Web services for process " + processDefinitionUUID + " successfully undeployed on server " + server + ".");
        } catch (final RuntimeException e) {
          // if exception
          if (ignoreExceptions) {
            ManagementAPIImpl.LOG.log(Level.WARNING,
                "Web services undeployment for process " + processDefinitionUUID + " failed on server " + server + ".", e);
          } else {
            throw e;
          }
        }
      }
    }
  }

  private void refreshProcessOnCluster(final ProcessDefinitionUUID processDefinitionUUID, final boolean ignoreExceptions) {
    // Get serverList
    final Collection<Server> servers = this.getOrchestraServersInCluster();
    // Deploy web services on each server
    for (final Server server : servers) {
      ManagementAPI managementAPI = null;
      if (server.equals(this.localServer)) {
        managementAPI = this;
      } else {
        // try to connect to server
        try {
          managementAPI = server.getManagementAPI();
        } catch (final Exception e) {
          // server is down, log a warning.
          ManagementAPIImpl.LOG.log(Level.WARNING, "Connection to server " + server + " failed.", e);
        }
      }
      // If server is up
      if (managementAPI != null) {
        try {
          // deploy web services on the server
          managementAPI.refreshProcess(processDefinitionUUID);
          ManagementAPIImpl.LOG.log(Level.INFO,
              "Process " + processDefinitionUUID + " successfully refreshed on server " + server + ".");
        } catch (final RuntimeException e) {
          // if exception
          if (ignoreExceptions) {
            ManagementAPIImpl.LOG.log(Level.WARNING,
                "Process " + processDefinitionUUID + " refresh failed on server " + server + ".", e);
          } else {
            throw e;
          }
        }
      }
    }
  }

  private Collection<Server> getOrchestraServersInCluster() {
    return this.execute(new Command<Collection<Server>>() {
      private static final long serialVersionUID = 1L;

      public Collection<Server> execute(final Environment environment) {
        final ClusterDescription clusterDescription = environment.get(ClusterDescription.class);
        if (clusterDescription != null) {
          return clusterDescription.getOrchestraServers();
        }
        return Collections.singletonList(ManagementAPIImpl.this.localServer);
      }
    });
  }


  public void deployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.execute(new DeployWSCommand(processDefinitionUUID));
  }

  public void undeployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.execute(new UndeployWSCommand(processDefinitionUUID));
  }

  public boolean undeploy(final QName processQName) {
    final ProcessDefinitionUUID processDefinitionUUID = this.execute(new GetActiveProcessDefinitionUUIDCommand(processQName));
    return this.undeploy(processDefinitionUUID);
  }

  @SuppressWarnings("boxing")
  public boolean undeploy(final ProcessDefinitionUUID pduuid) {
    final Set<ProcessDefinitionUUID> otherVersions = this.getOtherVersionUUIDs(pduuid);
    final boolean result;

    this.undeployWebServicesOnCluster(pduuid, false);
    try {
      result = this.getCommandService().execute(new UndeployCommand(pduuid));
    } catch (final RuntimeException e) {
      this.deployWebServicesOnCluster(pduuid, true);
      throw e;
    }
    for (final ProcessDefinitionUUID otherVersion : otherVersions) {
      this.refreshProcessOnCluster(otherVersion, true);
    }
    return result;
  }

  private Set<ProcessDefinitionUUID> getOtherVersionUUIDs(final ProcessDefinitionUUID pduuid) {
    final Set<ProcessDefinitionUUID> otherVersions;
    final ProcessDefinitionData pdata = this.getCommandService().execute(new GetProcessCommand(pduuid));
    final Set<ProcessDefinitionData> processes = this.getCommandService().execute(new GetProcessesCommand());
    otherVersions = new HashSet<ProcessDefinitionUUID>();
    for (final ProcessDefinitionData process : processes) {
      if (process.getQName().equals(pdata.getQName()) && !process.getUUID().equals(pduuid)) {
        otherVersions.add(process.getUUID());
      }
    }
    return otherVersions;
  }

  public boolean remove(final ProcessDefinitionUUID processDefinitionUUID) {
    return this.getCommandService().execute(new RemoveProcessCommand(processDefinitionUUID)).booleanValue();
  }

  public boolean activate(final ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException {
    try {
      final boolean result = this.getCommandService().execute(new ActivateProcessCommand(processDefinitionUUID)).booleanValue();
      if (result) {
        this.refreshProcessOnCluster(processDefinitionUUID, true);
      }
      return result;
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public boolean retire(final ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException {
    try {
      final boolean result = this.getCommandService().execute(new RetireProcessCommand(processDefinitionUUID)).booleanValue();
      if (result) {
        this.refreshProcessOnCluster(processDefinitionUUID, true);
      }
      return result;
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public void refreshProcess(final ProcessDefinitionUUID processDefinitionUUID) {
    this.getCommandService().execute(new RefreshProcessCommand(processDefinitionUUID));
  }

  public ProcessDefinition deployBar(final byte[] barFile) {
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      final File tmpDir = File.createTempFile("orch", null, null);
      tmpDir.delete();
      if (!tmpDir.mkdirs()) {
        throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
      }

      final List<URL> wsdlUrls = new ArrayList<URL>();
      URL bpelFile = null;

      // Unzip file
      final InputStream in = new ByteArrayInputStream(barFile);
      final ZipInputStream zipInput = new ZipInputStream(in);
      ZipEntry zipFile;

      final Map<String, byte[]> resources = new HashMap<String, byte[]>();

      while ((zipFile = zipInput.getNextEntry()) != null) {

        // write the file to the disk
        // Create file
        final File outFile = new File(tmpDir.getAbsolutePath() + File.separator + zipFile.getName());
        if (zipFile.isDirectory()) {
          outFile.mkdirs();
        } else {
          final byte[] content = Misc.getAllContentFrom(zipInput);
          resources.put(zipFile.getName(), content);
          Misc.write(content, outFile);// Copy file content

          if (zipFile.getName().endsWith(".wsdl")) {
            wsdlUrls.add(outFile.toURI().toURL());
          } else if (zipFile.getName().endsWith(".bpel")) {
            if (bpelFile != null) {
              throw new OrchestraRuntimeException("Bar contains more than one bpel file.");
            }
            bpelFile = outFile.toURI().toURL();
          }
        }
      }
      zipInput.close();
      in.close();

      final Deployment deployment = DeploymentUtil.generateDeployment(bpelFile, wsdlUrls);

      deployment.getResourcesRepository().getResources().putAll(resources);
      // Deploy the process
      final ProcessDefinition process = this.deploy(deployment);

      // Delete temp dir
      Misc.deleteDir(tmpDir);

      return process;
    } catch (final IOException e) {
      throw new OrchestraRuntimeException("Unable to deploy bar", e);
    }
  }

  public void initialize() {
    this.getCommandService().execute(new InitializeCommand());
  }

}
