/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.ow2.orchestra.cluster.Server;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.InstanceManagementAPI;
import org.ow2.orchestra.facade.QuerierDefinitionAPI;
import org.ow2.orchestra.facade.QuerierRuntimeAPI;
import org.ow2.orchestra.facade.data.runtime.JobData;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.exception.ActivityInstanceNotFoundException;
import org.ow2.orchestra.facade.exception.ActivityNotFoundException;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.JobNotFoundException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.cmd.Command;

public class RemoteDeployerImpl implements RemoteDeployerMBean {

  private final ManagementAPIImpl managementAPI;
  private final InstanceManagementAPI instanceManagementAPI;
  private final QuerierRuntimeAPI querierRuntimeAPI;
  private final QuerierDefinitionAPI querierDefinitionAPI;

  public RemoteDeployerImpl(final EnvironmentFactory environmentFactory) {
    this.managementAPI = new ManagementAPIImpl(environmentFactory);
    this.instanceManagementAPI = new InstanceManagementAPIImpl(environmentFactory);
    this.querierRuntimeAPI = new QuerierRuntimeAPIImpl(environmentFactory);
    this.querierDefinitionAPI = new QuerierDefinitionAPIImpl(environmentFactory);
  }


  public RemoteDeployerImpl(final Server orchestraServer, final EnvironmentFactory environmentFactory) {
    this.managementAPI = new ManagementAPIImpl(orchestraServer, environmentFactory);
    this.instanceManagementAPI = new InstanceManagementAPIImpl(environmentFactory);
    this.querierRuntimeAPI = new QuerierRuntimeAPIImpl(environmentFactory);
    this.querierDefinitionAPI = new QuerierDefinitionAPIImpl(environmentFactory);
  }

  // MANAGEMENT API
  public <T> T execute(final Command<T> command) {
    return this.managementAPI.execute(command);
  }

  public ProcessDefinition deploy(final Deployment bpelDeployment) {
    return this.managementAPI.deploy(bpelDeployment);
  }

  public void deployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.managementAPI.deployWebServices(processDefinitionUUID);
  }

  public void undeployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.managementAPI.undeployWebServices(processDefinitionUUID);
  }

  public boolean undeploy(final QName processQName) {
    return this.managementAPI.undeploy(processQName);
  }

  public boolean undeploy(final ProcessDefinitionUUID pduuid) {
    return this.managementAPI.undeploy(pduuid);
  }

  public boolean remove(final ProcessDefinitionUUID processDefinitionUUID) {
    return this.managementAPI.remove(processDefinitionUUID);
  }

  public boolean activate(final ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException {
    return this.managementAPI.activate(processDefinitionUUID);
  }

  public boolean retire(final ProcessDefinitionUUID processDefinitionUUID) throws ProcessNotFoundException {
    return this.managementAPI.retire(processDefinitionUUID);
  }

  public void refreshProcess(final ProcessDefinitionUUID processDefinitionUUID) {
    this.managementAPI.refreshProcess(processDefinitionUUID);
  }

  public ProcessDefinition deployBar(final byte[] barFile) {
    return this.managementAPI.deployBar(barFile);
  }

  public void initialize() {
    this.managementAPI.initialize();
  }

  // QUERIER DEFINITION API
  public List<ProcessDefinition> findProcessDefinitions() {
    return this.querierDefinitionAPI.findProcessDefinitions();
  }

  public List<ProcessDefinition> findProcessDefinitions(final ProcessState state) {
    return this.querierDefinitionAPI.findProcessDefinitions(state);
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName, final ProcessState state) {
    return this.querierDefinitionAPI.findProcessDefinitions(processName, state);
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName) {
    return this.querierDefinitionAPI.findProcessDefinitions(processName);
  }

  public ProcessDefinition getProcessDefinition(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    return this.querierDefinitionAPI.getProcessDefinition(processUUID);
  }

  public ActivityDefinition getActivityDefinition(final ActivityDefinitionUUID activityDefUUID) throws ActivityNotFoundException {
    return this.querierDefinitionAPI.getActivityDefinition(activityDefUUID);
  }

  public Set<ActivityDefinition> findActivityDefinitions(final ProcessDefinitionUUID processUUID)
    throws ProcessNotFoundException {
    return this.querierDefinitionAPI.findActivityDefinitions(processUUID);
  }

  // QUERIER RUNTIME API
  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    return this.querierRuntimeAPI.findActivityInstances(processUUID);
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID, final ActivityState activityState)
    throws InstanceNotFoundException {
    return this.querierRuntimeAPI.findActivityInstances(processUUID, activityState);
  }

  public Set<ProcessInstance> findProcessInstances() {
    return this.querierRuntimeAPI.findProcessInstances();
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID)
    throws ProcessNotFoundException {
    return this.querierRuntimeAPI.findProcessInstances(processDefinitionUUID);
  }

  public Set<ProcessInstance> findProcessInstances(
      final ProcessDefinitionUUID processDefinitionUUID,
      final ActivityState processInstanceState)
    throws ProcessNotFoundException {
    return this.querierRuntimeAPI.findProcessInstances(processDefinitionUUID, processInstanceState);
  }

  public Set<ProcessInstance> findProcessInstances(final ActivityState processInstanceState) {
    return this.querierRuntimeAPI.findProcessInstances(processInstanceState);
  }

  public ProcessInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    return this.querierRuntimeAPI.getProcessInstance(processInstanceUUID);
  }

  public ActivityInstance getActivityInstance(final ActivityInstanceUUID activityInstanceUUID)
    throws ActivityInstanceNotFoundException {
    return this.querierRuntimeAPI.getActivityInstance(activityInstanceUUID);
  }

  // INSTANCE MANAGEMENT API
  public void exit(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    this.instanceManagementAPI.exit(processInstanceUUID);
  }

  public void resume(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    this.instanceManagementAPI.resume(processInstanceUUID);
  }

  public void suspend(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    this.instanceManagementAPI.suspend(processInstanceUUID);
  }

  public boolean remove(final ProcessInstanceUUID processInstanceUUID) {
    return this.instanceManagementAPI.remove(processInstanceUUID);
  }

  public List<JobData> findJobsWithException() {
    return this.instanceManagementAPI.findJobsWithException();
  }

  public void retryJob(final JobData jobData, final int retries) throws JobNotFoundException {
    this.instanceManagementAPI.retryJob(jobData, retries);
  }
}
