package org.ow2.orchestra.osgi;

import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;

import javax.servlet.Servlet;
import javax.servlet.ServletException;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.ow2.orchestra.env.BpelEnvironmentParser;
import org.ow2.orchestra.jmx.JMXAgent;
import org.ow2.orchestra.osgi.impl.OrchestraExtensionClassLoader;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.env.PvmEnvironmentFactory;
import org.ow2.orchestra.services.itf.HttpPublisher;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

public class OrchestraOSGiEngine {

  public static final String ORCHESTRA_WAR_CONTEXT_PROP = "orchestra.war.context";
  public static final String ORCHESTRA_BUNDLE_CONTEXT_KEY = "orchestra.bundle.context";

  private static final class CreateEnvironmentClassLoaderAction implements PrivilegedAction<EnvironmentResourcesLoader> {

    private final ClassLoader oldClassLoader;
    private final Bundle envBundle;
    private final URL dirUrl;

    private CreateEnvironmentClassLoaderAction(final ClassLoader oldClassLoader, final URL dirUrl, final Bundle envBundle) {
      this.oldClassLoader = oldClassLoader;
      this.envBundle = envBundle;
      this.dirUrl = dirUrl;
    }
    public EnvironmentResourcesLoader run() {
      return new EnvironmentResourcesLoader(this.dirUrl, this.envBundle, this.oldClassLoader);
    }
  }

  private final Map<String, Servlet> aliases;

  // Injected by iPOJO
  private HttpService httpService = null;
  private final List<OrchestraExtensionService> orchestraExtensionServices = new ArrayList<OrchestraExtensionService>();
  private EnvironmentProviderService environmentProviderService;
  private Set<String> environmentDependencies = null;


  private JMXAgent jmxAgent;
  private EnvironmentFactory environmentFactory;

  private final ClassLoader bundleClassLoader  = new OrchestraExtensionClassLoader(this.orchestraExtensionServices);
  private final BundleContext bundleContext;

  private boolean canStart = false;

  public OrchestraOSGiEngine(final BundleContext bundleContext) {
    this.aliases = new HashMap<String, Servlet>();
    this.bundleContext = bundleContext;
  }

  public void setEnvironmentProviderService(final EnvironmentProviderService environmentProviderService) {
    this.environmentProviderService = environmentProviderService;
  }
  public EnvironmentProviderService getEnvironmentProviderService() {
    return this.environmentProviderService;
  }

  public synchronized HttpService getHttpService() {
    return this.httpService;
  }

  public synchronized void setHttpService(final HttpService httpService) {
    this.httpService = httpService;
  }

  public synchronized void unsetHttpService(final HttpService httpServiceToRemove) {
    // unregister servlets
    this.unregisterServlets();
    this.httpService = null;
  }


  private void unregisterServlets() {
    for (final String alias : this.aliases.keySet()) {
      this.httpService.unregister(alias);
      Misc.log(Level.INFO, "Alias unregistered: %s", alias);
    }
    this.aliases.clear();
  }


  /**
   * @param orchestraExtensionService the orchestraExtensionService to set
   */
  public synchronized void setOrchestraExtensionService(
      final OrchestraExtensionService orchestraExtensionService) throws Exception {
    this.orchestraExtensionServices.add(orchestraExtensionService);
    if (this.canStart) {
      try {
        this.startOrchestra();
      } catch (final Exception e) {
        e.printStackTrace();
        throw e;
      }

    }
  }

  public synchronized void unsetOrchestraExtensionService(
      final OrchestraExtensionService orchestraExtensionService) throws Throwable {
    this.orchestraExtensionServices.remove(orchestraExtensionService);

    // stop orchestra
    if (this.jmxAgent != null) {
      // check dependencies are still ok
      final Set<String> missingDependencies = this.checkDependencies();
      if (!missingDependencies.isEmpty()) {
        Misc.log(Level.SEVERE,
            "Orchestra will be stopped because some environment dependencies are no longer available: "
            + Misc.deepToString(missingDependencies));
        this.stop();
        this.canStart = true;
      }
    }
  }


  public synchronized void start() throws Exception {
    this.canStart = true;

    Misc.log(Level.FINEST, "%s starting", this.getClass().getName());

    // open environment
    Misc.log(Level.INFO, "Environment will be taken from: %s", this.environmentProviderService.getEnvironmentURL());

    try {
      // parse classes
      this.environmentDependencies = new HashSet<String>();
      final Document doc = XmlUtil.getDocumentFromURL(this.environmentProviderService.getEnvironmentURL());
      final NodeList nl =  doc.getElementsByTagName("*");
      for (int i = 0; i < nl.getLength(); i++) {
        final Element n = (Element) nl.item(i);
        final String className = XmlUtil.attribute(n, "class");
        if (className != null) {
          this.environmentDependencies.add(className);
        }
      }
      this.startOrchestra();
    } catch (final Exception t) {
      t.printStackTrace();
      throw t;
    }
  }

  private Set<String> checkDependencies() {
    // check dependencies are ok
    final Set<String> missingDependencies = new HashSet<String>();
    for (final String className : this.environmentDependencies) {
      try {
        this.bundleClassLoader.loadClass(className);
      } catch (final ClassNotFoundException e) {
        missingDependencies.add(className);
      }
    }
    return missingDependencies;
  }

  private synchronized void startOrchestra() throws ServletException, NamespaceException {
    final ClassLoader originalClassLoader = Thread.currentThread().getContextClassLoader();
    Thread.currentThread().setContextClassLoader(this.bundleClassLoader);
    try {
      // check dependencies are ok
      final Set<String> missingDependencies = this.checkDependencies();
      if (!missingDependencies.isEmpty()) {
        Misc.log(Level.SEVERE, "Some environment dependencies are missing: " + Misc.deepToString(missingDependencies));
        return;
      }

      final ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
      ClassLoader newClassLoader = oldClassLoader;
      URL dirUrl = null;
      try {
        dirUrl = this.environmentProviderService.getEnvironmentURL().toURI().resolve("./").toURL();
      } catch (final MalformedURLException e) {
        Misc.unreachableStatement();
      } catch (final URISyntaxException e) {
        Misc.unreachableStatement();
      }
      if (dirUrl != null) {
        newClassLoader = AccessController.doPrivileged(
            new CreateEnvironmentClassLoaderAction(oldClassLoader, dirUrl,
                this.environmentProviderService.getEnvironmentBundle()));
      }
      Thread.currentThread().setContextClassLoader(newClassLoader);
      try {
        this.environmentFactory =
          BpelEnvironmentParser.parseEnvironmentFactoryFromURL(this.environmentProviderService.getEnvironmentURL());
        ((PvmEnvironmentFactory)this.environmentFactory).getEnvironmentFactoryCtxWireContext().setClassLoader(newClassLoader);
        ((PvmEnvironmentFactory)this.environmentFactory).getEnvironmentCtxWireDefinition().setClassLoader(newClassLoader);
      } finally {
        Thread.currentThread().setContextClassLoader(oldClassLoader);
      }

      // Add bundle context to environment
      this.environmentFactory.set(OrchestraOSGiEngine.ORCHESTRA_BUNDLE_CONTEXT_KEY, this.bundleContext);

      // register cxfServlet
      // get web app path from orchestra.properties
      final Properties orchestraProperties =
        (Properties) this.environmentFactory.get(OrchestraConstants.ORCHESTRA_PROPERTIES);
      final String servletPath = orchestraProperties.getProperty(OrchestraConstants.SERVLET_PATH);

      final HttpContext context = this.httpService.createDefaultHttpContext();
      Misc.log(Level.FINEST, "Context created: " + context);

      final HttpPublisher publisher = this.environmentFactory.get(HttpPublisher.class);
      final Servlet servlet = publisher.getPublisherServlet(orchestraProperties);
      Misc.log(Level.FINEST, "Servlet created: %s", servlet);
      final Dictionary<String, String> initparams = new Hashtable<String, String>();


      final String inWar = this.bundleContext.getProperty(OrchestraOSGiEngine.ORCHESTRA_WAR_CONTEXT_PROP);
      if (inWar != null && Boolean.TRUE.equals(XmlUtil.parseBooleanValue(inWar))) {
        this.aliases.put("/", servlet);
      } else {
        this.aliases.put("/" + servletPath, servlet);
      }

      for (final Map.Entry<String, Servlet> mapping : this.aliases.entrySet()) {
        this.httpService.registerServlet(mapping.getKey(), mapping.getValue(), initparams, context);
        Misc.log(Level.INFO, "Alias registered: %s for servlet: %s", mapping.getKey(), mapping.getValue());
      }

      // create jmxAgent
      this.jmxAgent = new JMXAgent();

      // start orchestra
      this.jmxAgent.startOrchestra(this.environmentFactory);

      this.canStart = false;
    } finally {
      Thread.currentThread().setContextClassLoader(originalClassLoader);
    }
  }

  public synchronized void stop() throws Throwable {
    this.canStart = false;
    final ClassLoader originalClassLoader = Thread.currentThread().getContextClassLoader();
    if (this.bundleClassLoader != null) {
      Thread.currentThread().setContextClassLoader(this.bundleClassLoader);
    }
    try {
      Misc.log(Level.FINEST, "%s stopping", this.getClass().getName());

      // stop orchestra
      if (this.httpService != null) {
        this.unregisterServlets();
      }
      if (this.jmxAgent != null) {
        this.jmxAgent.stopOrchestra();
        this.jmxAgent = null;
        this.environmentFactory = null;
      }
    } catch (final Throwable t) {
      t.printStackTrace();
      throw t;
    } finally {
      Thread.currentThread().setContextClassLoader(originalClassLoader);
    }
  }
}
