/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.persistence.db.hibernate;

import java.io.IOException;
import java.io.Reader;
import java.io.Serializable;
import java.io.StringReader;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import org.hibernate.usertype.UserType;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Maps an XML document to a {@link java.sql.Clob}
 *
 * @author Juan Cantu
 * @author Guillaume Porcher
 */
public class DocumentType implements UserType {


  static final int[] SQL_TYPES = { Types.CLOB };

  public boolean equals(final Object x, final Object y) {
    if (x == null) {
      return y == null;
    }
    if (!((x instanceof Document) && y instanceof Document)) {
      return x.equals(y);
    }
    return XmlUtil.toString((Document) x).equals(XmlUtil.toString((Document) y));
  }

  public int hashCode(final Object x) {
    return x.hashCode();
  }

  public Object deepCopy(final Object value) {
    if (value != null) {
      return XmlUtil.getDocumentFromString(XmlUtil.toString((Node) value));
    }
    return null;
  }

  public boolean isMutable() {
    return true;
  }

  public Serializable disassemble(final Object value) {
    if (value != null) {
      return XmlUtil.toString((Node) value);
    }
    return null;
  }

  public Object assemble(final Serializable cached, final Object owner) {
    if (cached == null) {
      return null;
    }
    // parse XML text
    return XmlUtil.getDocumentFromString((String) cached);
  }

  public Object replace(final Object original, final Object target, final Object owner) {
    return this.deepCopy(original);
  }

  public int[] sqlTypes() {
    return DocumentType.SQL_TYPES;
  }

  public Object nullSafeGet(final ResultSet rs, final String[] names, final Object owner) throws SQLException {
    // retrieve stream from database
    final String columnName = names[0];
    Reader reader = null;

    try {
      reader = rs.getCharacterStream(columnName);
      // if SQL value is NULL, element is null as well
      if (reader == null) {
        return null;
      }

      final byte[] bytes = Misc.getAllContentFrom(reader);
      return this.assemble(new String(bytes), owner);

    } catch (final IOException e) {
      throw new OrchestraRuntimeException(e);
    } finally {
      if (reader != null) {
        try {
          reader.close();
        } catch (final Exception e) {
          // nothing
        }
      }
    }
  }

  public void nullSafeSet(final PreparedStatement st, final Object value, final int index) throws SQLException {
    // easy way out: null value
    if (value == null) {
      st.setNull(index, DocumentType.SQL_TYPES[0]);
    } else {
      final Node node = (Node) value;
      final String valueAsString = XmlUtil.toString(node);
      final Reader sr = new StringReader(valueAsString);
      st.setCharacterStream(index, sr, valueAsString.length());
    }
  }

  public Class<Document> returnedClass() {
    return Document.class;
  }
}


