package org.ow2.orchestra.services;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.IAbstractActivity;
import org.ow2.orchestra.definition.activity.IAbstractBpelActivity;
import org.ow2.orchestra.definition.activity.IScope;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.services.impl.ReplierImpl;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
/**
 * base implementation for the services to be published for a deployed bpel
 * process.
 *
 * @author Tom Baeyens, Charles Souillard, Guillaume Porcher
 */
public class Receiver {

  /** log. */
  private static Logger log = Logger.getLogger(Receiver.class.getName());


  protected Receiver() {
    super();
  }

  /**
   * Find if there is a stored message for the execution. If a message is found,
   * associate the stored message to this execution, and
   * return this execution.
   *
   * @param execution
   *          execution to check
   * @return matching message.
   */
  static ExecElementToSignal findMessageForExecution(final BpelExecution execution) {
    if (execution == null || execution.isFinished() || !execution.hasWaitingExecution()) {
      return null;
    }
    final WaitingExecution waitingExecution = execution.getWaitingExecution();
    for (final ReceivingElement receivingElement : waitingExecution.getInboundMessageElement().getReceivingElements()) {
      final PendingMessage pendingMessage =
        Receiver.getMessageFromStoredMessage(receivingElement, waitingExecution, execution.getProcessDefinition());
      if (pendingMessage != null) {
        Receiver.removePendingMessage(pendingMessage);
        return new ExecElementToSignal(execution, receivingElement, pendingMessage);
      }
    }
    return null;
  }

  static ExecElementToSignal associatePendingMessage(
      final PendingMessage pendingMessage,
      final BpelProcess bpelProcess,
      final OperationKey operationKey,
      final Date parentInstanceStartTime) {
    ExecElementToSignal execElementToSignal = null;

    final InboundMessageElement startActivity = bpelProcess.getStartElement(operationKey);
    final List<ReceivingElement> receivingElements = bpelProcess.getReceivingElements(operationKey);


    if (startActivity == null || bpelProcess.getNumberOfStartElements() > 1 || receivingElements.size() > 1) {
      // find a waiting execution using the correlations
      // skip if this is the only start activity
      execElementToSignal = Receiver.getWaitingExecution(pendingMessage, receivingElements, bpelProcess);
    }
    // if an execution was waiting for such an incoming message
    if (execElementToSignal != null) {
      if (Receiver.log.isLoggable(Level.FINE)) {
        Receiver.log.fine("Matching execution found for incoming msg: " + execElementToSignal.getBpelExecution());
      }

      return execElementToSignal;
    } else {
      // no execution was found
      if (Receiver.log.isLoggable(Level.FINE)) {
        Receiver.log.fine("no matching execution found for pending msg.");
      }
      return null;
    }
  }


  static ExecElementToSignal startNewInstance(
      final PendingMessage pendingMessage,
      final BpelProcess bpelProcess,
      final OperationKey operationKey,
      final Date parentInstanceStartTime) {
    final ExecElementToSignal execElementToSignal = null;

    final InboundMessageElement startActivity = bpelProcess.getStartElement(operationKey);
    // if the incoming message matched a start activity
    if (startActivity != null) {
      ReceivingElement startElement = null;
      for (final ReceivingElement receivingElement : startActivity.getReceivingElements()) {
        if (receivingElement.getOperationKey().equals(operationKey)) {
          startElement = receivingElement;
        }
      }
      if (startElement == null) {
        throw new IllegalStateException("No start element found");
      }
      if (Receiver.log.isLoggable(Level.FINE)) {
        Receiver.log.fine("found start node " + startElement + " for incoming msg");
      }
      // create new instance
      final ExecElementToSignal receiveExecution =
        Receiver.createNewInstance(pendingMessage, bpelProcess, startActivity, startElement, parentInstanceStartTime);

      if (receiveExecution == null) {
        throw new OrchestraRuntimeException("No waiting execution found !, process = "
          + bpelProcess.getQName());
      }
      return receiveExecution;
    } else { // no start node was found
      return null;
    }
  }

  private static ExecElementToSignal createNewInstance(
      final PendingMessage pendingMessage,
      final BpelProcess bpelProcess,
      final InboundMessageElement startActivity,
      final ReceivingElement startElement,
      final Date parentInstanceStartTime) {
    // create instance
    final BpelExecution bpelExecution = (BpelExecution) bpelProcess.createProcessInstance();
    // initialise receive variables which are process variables
    // (need to be initialized for process event handlers
    final Map<String, Object> additionnalVariables = new HashMap<String, Object>();
    // Get process scope (casting to process causes a bug with hibernate proxies)
    final IScope processActivity = (IScope) bpelProcess.getInitial().getBehaviour();
    if (startElement.getVariable() != null) {
      final Variable var =
        ((IAbstractBpelActivity) startActivity).getEnclosingScope().findVariable(startElement.getVariable());
      final Variable var1 = processActivity.getVariable(startElement.getVariable());
      if (var1.equals(var)) {
        additionnalVariables.put(startElement.getVariable(), pendingMessage.getMessage().duplicate());
      }
    } else {
      for (final Copy copy : startElement.getInCopies()) {
        final Variable var =
          ((IAbstractBpelActivity) startActivity).getEnclosingScope().findVariable(copy.getTo().getVariable());
        final Variable var1 = processActivity.getVariable(copy.getTo().getVariable());
        if (var1.equals(var)) {
          additionnalVariables.put(copy.getTo().getVariable(),
              pendingMessage.getMessage().getPartValue(copy.getFrom().getPart()));
        }
      }
    }
    if (!additionnalVariables.isEmpty()) {
      bpelExecution.createVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES, additionnalVariables);
    }
    if (parentInstanceStartTime != null) {
      bpelExecution.setParentInstanceStartTime(parentInstanceStartTime);
    }
    // start execution
    bpelExecution.begin();

    return Receiver.getStartElementExecution(bpelExecution, startElement, pendingMessage);
  }

  /**
   * Associate the given replier to the corresponding runtimeScope (top level
   * scope if messageExchange is not specified, scopeRuntime declaring the used
   * messageExchange if there is one.
   */
  public static void addReplier(final BpelExecution bpelExecution,
      final ReceivingElement receivingElement,
      final MessageCarrier messageCarrier) {
    if (messageCarrier != null) {
      final String messageExchange = receivingElement.getMessageExchange();
      final ReplierKey replierKey =
        new ReplierKey(receivingElement.getOperationKey(), messageExchange, bpelExecution.getCurrentScopeRuntime());
      final Replier replier = Receiver.createReplier(replierKey);
      MessageCarrierRepository.addMessageCarrier(replierKey, messageCarrier);
      bpelExecution.addReplier(replier);
    }
  }

  /**
   * Search for an execution of the given process waiting for a message on the
   * given portType/operation. I one was found and the cs matches (if any),
   * associate this execution to the given message and return this execution.
   * @param operationKey
   *
   * @param incomingMessage
   *          message received we want to give to an execution
   * @param operationKey
   *          operationKey
   * @param environment
   *          environment
   * @return the found execution or null if no one was found
   */
  private static ExecElementToSignal getWaitingExecution(
      final PendingMessage pendingMessage,
      final List<ReceivingElement> receivingElements,
      final BpelProcess bpelProcess) {

    final MessageVariable incomingMessage = pendingMessage.getMessage();

    for (final ReceivingElement receivingElement : receivingElements) {
      if (receivingElement.getCorrelations() == null || receivingElement.getCorrelations().isEmpty()) {
        // Choose first available waitingExecution
        final WaitingExecution waitingExecution =
          EnvTool.getRepository().getFirstWaitingExecution(receivingElement.getInboundMessageElement());
        if (waitingExecution != null) {
          final ExecElementToSignal res =
            new ExecElementToSignal(waitingExecution.getBpelExecution(), receivingElement, pendingMessage);
          return res;
        }
      } else {
        final Map<QName, Document> messageProperties =
          Receiver.getCSPropertiesFromMessage(incomingMessage, receivingElement, bpelProcess);
        final ResultIterator<WaitingExecution> waitingExecutions =
          EnvTool.getRepository().getWaitingExecutions(receivingElement, messageProperties);
        if (Receiver.log.isLoggable(Level.FINE)) {
          Receiver.log.fine("Waiting executions for receiving element " + receivingElement + ": " + waitingExecutions);
        }
        WaitingExecution waitingExecution = waitingExecutions.next();
        while (waitingExecution != null) {
          if (Receiver.csMatch(incomingMessage, receivingElement, waitingExecution, bpelProcess)) {
            final ExecElementToSignal res =
              new ExecElementToSignal(waitingExecution.getBpelExecution(), receivingElement, pendingMessage);
            return res;
          }
          waitingExecution = waitingExecutions.next();
        }
      }
    }
    if (Receiver.log.isLoggable(Level.FINE)) {
      Receiver.log.fine("No waiting executions for Receiver: " + pendingMessage.getOperationKey());
    }
    return null;
  }

  /**
   * Search for an execution of the given process instance waiting for a message
   * on the given node. This method is used to associate an incoming message
   * which creates a new instance to the newly created instance.
   *
   * @return the found execution or null if no one was found
   */
  private static ExecElementToSignal getStartElementExecution(final BpelExecution parent, final ReceivingElement startElement,
    final PendingMessage pendingMessage) {

    if (parent == null) {
      return null;
    }
    final BpelExecution instance = parent.getProcessInstance();
    final BpelProcess bpelProcess = instance.getProcessDefinition();

    final Set<BpelExecution> waitingExec = instance.searchWaitingExecutions();
    if (Receiver.log.isLoggable(Level.FINE)) {
      Receiver.log.fine(Receiver.class + ".getStartElementExecution, execution = " + parent + " waiting=" + waitingExec);
    }
    final ExecElementToSignal res = null;
    if (waitingExec != null) {
      for (final BpelExecution child : waitingExec) {
        final IAbstractActivity activity = (IAbstractActivity) child.getNode().getBehaviour();
        if (activity.isInboundMessageElement()) {
          // Add execution to receiver if it is not the right execution
          if (((InboundMessageElement) activity).getReceivingElements().contains(startElement)) {
            if (bpelProcess.getCommonCorrelationSets() != null) {
              for (final CorrelationSet cs : bpelProcess.getCommonCorrelationSets()) {
                ReceivingElementUtil.initiateCs(startElement, child, cs, pendingMessage.getMessage());
              }
            }
            return new ExecElementToSignal(child, startElement, pendingMessage, true);
          }
        }
      }
    }
    return res;
  }

  /**
   * Store the given message as a message that could be consumed by an activity
   * waiting for a message on the given portType/operation.
   *
   * @param message
   *          message to store
   * @param operationKey
   * @param operationName
   *          operation
   * @param portType
   *          portType
   * @param environment
   *          environment
   *
   */
  static PendingMessage storeIncomingMessage(final MessageVariable message,
      final MessageCarrier messageCarrier, final OperationKey operationKey) {
    if (Receiver.log.isLoggable(Level.FINEST)) {
      Receiver.log.finest("Incoming msg: " + message);
    }
    PendingMessage pendingMessage = null;
    if (messageCarrier != null) {
      MessageCarrierRepository.addPendingMessageCarrier(messageCarrier);
      pendingMessage = new PendingMessage(message, messageCarrier.getUuid(), operationKey);
    } else {
      pendingMessage = new PendingMessage(message, null, operationKey);
    }
    return pendingMessage;
  }

  protected static void storePendingMessage(final PendingMessage pendingMessage) {
    EnvTool.getRepository().storePendingMessage(pendingMessage);
  }
  // ///////////////////////////////////////////////////////////////////////////////////////

  /**
   * Return true if the given message match the given startActivity for the
   * given execution. It takes in charge cs verification.
   *
   * @param messageToAnalyse
   *          messageToAnalyse
   * @param receivingElement
   *          startActivity
   * @param waitingExecution
   *          bpelExecution
   * @return true if the given message match the given startActivity for the
   *         given execution.
   */
  static boolean csMatch(
      final MessageVariable messageToAnalyse,
      final ReceivingElement receivingElement,
      final WaitingExecution waitingExecution,
      final BpelProcess bpelProcess) {
    if (receivingElement.getCorrelations() == null) {
      return true;
    }
    // we calcul the values for each correlationSet
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (!correlation.correlationOK(messageToAnalyse, receivingElement, waitingExecution, bpelProcess)) {
        return false;
      }
    }
    return true;

  }

  private static Map<QName, Document> getCSPropertiesFromMessage(
      final MessageVariable message,
      final ReceivingElement receivingElement,
      final BpelProcess bpelProcess) {
    if (receivingElement.getCorrelations() == null) {
      return Collections.emptyMap();
    }
    final Map<QName, Document> properties = new HashMap<QName, Document>();

    final QName messageType = receivingElement.getInputMessageType();
    // we compute the values for each correlationSet
    for (final Correlation correlation : receivingElement.getCorrelations()) {
      if (correlation.getInitiate().equals(Initiate.no) || correlation.getInitiate().equals(Initiate.join)) {
        final List<PropertyAlias> propertyAliases =
          correlation.getCorrelationSet().getPropertyAliasRefs(bpelProcess.getWsdlInfos(), messageType);
        for (final PropertyAlias propertyAlias : propertyAliases) {
          final Node value = propertyAlias.getPropertyValueFromMessage(message);
          final Element el = XmlUtil.getDocumentWithOneElement(
              new QName(XmlConstants.ELT_DEFAULT_NS, XmlConstants.ELT_DEFAULT_ELEMENT));
          el.appendChild(el.getOwnerDocument().importNode(value, true));
          properties.put(propertyAlias.getPropertyName(), el.getOwnerDocument());
        }
      }
    }
    return properties;
  }

  /**
   * Search for a stored message matching the given parameters.
   *
   * @param bpelExecution
   *          bpelExecution
   * @param receivingElement
   *          startActivity
   * @return a stored message matching the given parameters. or null.
   */
  private static PendingMessage getMessageFromStoredMessage(
      final ReceivingElement receivingElement,
      final WaitingExecution waitingExecution,
      final BpelProcess bpelProcess) {
    final ResultIterator<PendingMessage> pendingMessages =
      EnvTool.getRepository().getPendingMessages(receivingElement.getOperationKey());
    PendingMessage pendingMessage = pendingMessages.next();
    while (pendingMessage != null) {
      final MessageVariable message = pendingMessage.getMessage();
      if (Receiver.csMatch(message, receivingElement, waitingExecution, bpelProcess)) {
        return pendingMessage;
      }
      pendingMessage = pendingMessages.next();
    }
    return null;
  }

  /**
   * Remove pending message from repository. The pending message can be given to an execution.
   * @param pendingMessage
   */
  public static void removePendingMessage(final PendingMessage pendingMessage) {
    EnvTool.getRepository().removePendingMessage(pendingMessage);
  }


  /**
   * Delete pending message from repository. The pending message is not given to an execution.
   * The associated messageCarrier should be notified that no answer will be available.
   * @param pendingMessage
   */
  public static void deletePendingMessage(final PendingMessage pendingMessage) {
    EnvTool.getRepository().removePendingMessage(pendingMessage);
    if (pendingMessage.getMessageCarrierId() != null) {
      MessageCarrierRepository.throwMissingReply(pendingMessage.getMessageCarrierId());
    }

  }

  private static Replier createReplier(final ReplierKey replierKey) {
    return new ReplierImpl(replierKey);
  }

}
