/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.wsdl;

import java.util.List;
import java.util.Set;

import javax.wsdl.Binding;
import javax.wsdl.BindingOperation;
import javax.wsdl.Message;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPOperation;

import org.ow2.orchestra.facade.exception.IncompatibleWSDLException;


/**
 * @author Guillaume Porcher
 *
 */
public class WsdlComparator {

  /**
   * Compare the two ports. Throw {@link IncompatibleWSDLException} if they are not the same.
   * Does not compare the SOAP Address part (can be overridden by WS engines during deployment).
   * @param oldPort
   * @param newPort
   */
  public static void checkSamePort(final Port oldPort, final Port newPort) {
    // compare names
    if (!WsdlComparator.nullSafeEquals(oldPort.getName(),newPort.getName())) {
      throw new IncompatibleWSDLException("Port name changed: was " + oldPort.getName() + ", now is " + newPort.getName());
    }
    // don't compare SOAP addresses
    // because it can be modified later by WS engines.

    // compare binding
    WsdlComparator.checkSameBinding(oldPort.getBinding(), newPort.getBinding());
  }

  private static void checkSameBinding(final Binding oldBinding, final Binding newBinding) {
    // compare names
    if (!WsdlComparator.nullSafeEquals(oldBinding.getQName(), newBinding.getQName())) {
      throw new IncompatibleWSDLException(
          "Binding name changed: was " + oldBinding.getQName() + ", now is " + newBinding.getQName());
    }
    // compare SOAP addresses
    WsdlComparator.checkSameSOAPBinding(oldBinding, newBinding);

    // compare binding for operations
    if (oldBinding.getBindingOperations().size() != newBinding.getBindingOperations().size()) {
      throw new IncompatibleWSDLException(
          "Binding operation number changed: was " + oldBinding.getBindingOperations().size()
          + ", now is " + newBinding.getBindingOperations().size());
    }
    for (final BindingOperation oldBindingOp : (List<BindingOperation>) oldBinding.getBindingOperations()) {
      final BindingOperation newBindingOp = newBinding.getBindingOperation(
          oldBindingOp.getName(),
          oldBindingOp.getBindingInput() != null ? oldBindingOp.getBindingInput().getName() : null,
          oldBindingOp.getBindingOutput() != null ? oldBindingOp.getBindingOutput().getName() : null);
      if (newBindingOp == null) {
        throw new IncompatibleWSDLException(
            "Binding operation " + oldBindingOp.getName() + " not found.");
      }
      WsdlComparator.checkSameOperationBinding(oldBindingOp, newBindingOp);
    }
    // compare portType
    WsdlComparator.checkSamePortType(oldBinding.getPortType(), newBinding.getPortType());
  }

  private static void checkSameOperationBinding(final BindingOperation oldBindingOp, final BindingOperation newBindingOp) {
    if ((oldBindingOp.getBindingInput() != null && newBindingOp.getBindingInput() == null)
        || (oldBindingOp.getBindingInput() == null && newBindingOp.getBindingInput() != null)) {
      throw new IncompatibleWSDLException(
            "Input binding for operation " + oldBindingOp.getName() + " not found.");
    }
    if (oldBindingOp.getBindingInput() != null) {
      WsdlComparator.checkSameSOAPBody(
          oldBindingOp.getBindingInput().getExtensibilityElements(),
          newBindingOp.getBindingInput().getExtensibilityElements());
    }

    if ((oldBindingOp.getBindingOutput() != null && newBindingOp.getBindingOutput() == null)
        || (oldBindingOp.getBindingOutput() == null && newBindingOp.getBindingOutput() != null)) {
      throw new IncompatibleWSDLException(
            "Outut binding for operation " + oldBindingOp.getName() + " not found.");
    }
    if (oldBindingOp.getBindingOutput() != null) {
      WsdlComparator.checkSameSOAPBody(
          oldBindingOp.getBindingOutput().getExtensibilityElements(),
          newBindingOp.getBindingOutput().getExtensibilityElements());
    }

    WsdlComparator.checkSameSOAPOperation(oldBindingOp, newBindingOp);

    // TODO faults
  }

  private static void checkSameSOAPBody(
      final List<ExtensibilityElement> oldExtElements,
      final List<ExtensibilityElement> newExtElements) {
    SOAPBody oldSoapBody = null;
    SOAPBody newSoapBody = null;
    for (final ExtensibilityElement oldExtElt : oldExtElements) {
      if (oldExtElt instanceof SOAPBody) {
        oldSoapBody = (SOAPBody) oldExtElt;
      }
    }
    for (final ExtensibilityElement newExtElt : newExtElements) {
      if (newExtElt instanceof SOAPBody) {
        newSoapBody = (SOAPBody) newExtElt;
      }
    }
    if ((oldSoapBody == null && newSoapBody != null)
        || (oldSoapBody != null && newSoapBody == null)
        || (oldSoapBody != null && newSoapBody != null
            && !(WsdlComparator.nullSafeEquals(oldSoapBody.getEncodingStyles(), newSoapBody.getEncodingStyles())
                && WsdlComparator.nullSafeEquals(oldSoapBody.getUse(), newSoapBody.getUse())
                && WsdlComparator.nullSafeEquals(oldSoapBody.getParts(), newSoapBody.getParts())
                && WsdlComparator.nullSafeEquals(oldSoapBody.getNamespaceURI(), newSoapBody.getNamespaceURI())))) {
      throw new IncompatibleWSDLException("Soap body changed.");
    }
  }

  private static void checkSameSOAPOperation(final BindingOperation oldBindingOp, final BindingOperation newBindingOp) {
    SOAPOperation oldSoapOperation = null;
    SOAPOperation newSoapOperation = null;
    for (final ExtensibilityElement oldExtElt : (List<ExtensibilityElement>) oldBindingOp.getExtensibilityElements()) {
      if (oldExtElt instanceof SOAPOperation) {
        oldSoapOperation = (SOAPOperation) oldExtElt;
      }
    }
    for (final ExtensibilityElement newExtElt : (List<ExtensibilityElement>) newBindingOp.getExtensibilityElements()) {
      if (newExtElt instanceof SOAPOperation) {
        newSoapOperation = (SOAPOperation) newExtElt;
      }
    }
    if ((oldSoapOperation == null && newSoapOperation != null)
        || (oldSoapOperation != null && newSoapOperation == null)
        || (oldSoapOperation != null && newSoapOperation != null
            && !(WsdlComparator.nullSafeEquals(oldSoapOperation.getStyle(), newSoapOperation.getStyle())
                && WsdlComparator.nullSafeEquals(oldSoapOperation.getSoapActionURI(), newSoapOperation.getSoapActionURI())))) {
      throw new IncompatibleWSDLException("Soap operation " + oldBindingOp.getName() + " changed.");
    }
  }

  private static void checkSameSOAPBinding(final Binding oldBinding, final Binding newBinding) {
    SOAPBinding oldSoapBinding = null;
    SOAPBinding newSoapBinding = null;
    for (final ExtensibilityElement oldExtElt : (List<ExtensibilityElement>) oldBinding.getExtensibilityElements()) {
      if (oldExtElt instanceof SOAPBinding) {
        oldSoapBinding = (SOAPBinding) oldExtElt;
      }
    }
    for (final ExtensibilityElement newExtElt : (List<ExtensibilityElement>) newBinding.getExtensibilityElements()) {
      if (newExtElt instanceof SOAPBinding) {
        newSoapBinding = (SOAPBinding) newExtElt;
      }
    }
    if ((oldSoapBinding == null && newSoapBinding != null)
        || (oldSoapBinding != null && newSoapBinding == null)
        || (oldSoapBinding != null && newSoapBinding != null
            && !(WsdlComparator.nullSafeEquals(oldSoapBinding.getStyle(), newSoapBinding.getStyle())
                && WsdlComparator.nullSafeEquals(oldSoapBinding.getTransportURI(), newSoapBinding.getTransportURI())))) {
      throw new IncompatibleWSDLException("Soap binding changed.");
    }
  }

  private static void checkSamePortType(final PortType oldPortType, final PortType newPortType) {
    // compare names
    if (!WsdlComparator.nullSafeEquals(oldPortType.getQName(), newPortType.getQName())) {
      throw new IncompatibleWSDLException(
          "Binding name changed: was " + oldPortType.getQName() + ", now is " + newPortType.getQName());
    }

    // compare operations
    if (oldPortType.getOperations().size() != newPortType.getOperations().size()) {
      throw new IncompatibleWSDLException(
          "PortType operation number changed: was " + oldPortType.getOperations().size()
          + ", now is " + newPortType.getOperations().size());
    }
    for (final Operation oldOp : (List<Operation>) oldPortType.getOperations()) {
      final Operation newOp = newPortType.getOperation(
          oldOp.getName(),
          oldOp.getInput() != null ? oldOp.getInput().getName() : null,
              oldOp.getOutput() != null ? oldOp.getOutput().getName() : null);
      if (newOp == null) {
        throw new IncompatibleWSDLException(
            "PortType operation " + oldOp.getName() + " not found.");
      }
      WsdlComparator.checkSameOperation(oldOp, newOp);
    }
  }

  private static void checkSameOperation(final Operation oldOp, final Operation newOp) {
    if ((oldOp.getInput() != null && newOp.getInput() == null)
        || (oldOp.getInput() == null && newOp.getInput() != null)) {
      throw new IncompatibleWSDLException(
            "Input for operation " + oldOp.getName() + " not found.");
    }
    if (oldOp.getInput() != null) {
      WsdlComparator.checkSameMessage(oldOp.getInput().getMessage(), newOp.getInput().getMessage());
    }
    if ((oldOp.getOutput() != null && newOp.getOutput() == null)
        || (oldOp.getOutput() == null && newOp.getOutput() != null)) {
      throw new IncompatibleWSDLException(
            "Outut binding for operation " + oldOp.getName() + " not found.");
    }
    if (oldOp.getOutput() != null) {
      WsdlComparator.checkSameMessage(oldOp.getOutput().getMessage(), newOp.getOutput().getMessage());
    }

    // TODO Faults
  }

  private static void checkSameMessage(final Message oldMsg, final Message newMsg) {
    // compare names
    if (!WsdlComparator.nullSafeEquals(oldMsg.getQName(), newMsg.getQName())) {
      throw new IncompatibleWSDLException(
          "Message name changed: was " + oldMsg.getQName() + ", now is " + newMsg.getQName());
    }
    if (oldMsg.getParts().size() != newMsg.getParts().size()) {
      throw new IncompatibleWSDLException(
          "Message part number changed: was " + oldMsg.getParts().size() + ", now is " + newMsg.getParts().size());
    }
    final Set<String> partNames = oldMsg.getParts().keySet();
    for (final String partName : partNames) {
      final Part oldPart = oldMsg.getPart(partName);
      final Part newPart = newMsg.getPart(partName);
      if (newPart == null) {
        throw new IncompatibleWSDLException(
            "Part " + partName + " for message " + oldMsg.getQName() + " not found.");
      }
      if (!WsdlComparator.nullSafeEquals(oldPart.getTypeName(), newPart.getTypeName())) {
        throw new IncompatibleWSDLException(
            "Part " + partName + " type changed: was " + oldPart.getTypeName() + ", now is " + newPart.getTypeName());
      }
      if (!WsdlComparator.nullSafeEquals(oldPart.getElementName(), newPart.getElementName())) {
        throw new IncompatibleWSDLException(
            "Part " + partName + " element changed: was " + oldPart.getElementName() + ", now is " + newPart.getElementName());
      }
    }
  }

  private static boolean nullSafeEquals(final Object obj1, final Object obj2) {
    if (obj1 == null) {
      return obj2 == null;
    }
    return obj1.equals(obj2);
  }
}
